package MCUtils::MCWS;

use diagnostics;
use v5.14;
use utf8;
use strict;
use warnings;

my $version = "1.02";

use MCUtils::Misc;
use URI;
use URI::Escape;
use HTML::Entities qw(decode_entities);

=pod
    Constructor
	my $mcws = MCWS->connect(%mcwsparams);
=cut

sub connect {
    my $invocant = shift;
    my $class = ref($invocant) || $invocant;
    my $self = {
	 ua		=> LWP::UserAgent->new(),
	 timeout	=> 10,
	 agent		=> 'Mozilla/5.0',
	 server		=> 'localhost:52199',
	 realm		=> 'Library Server',
	 @_,
    };

    for (qw(username password)) {
	exists $self->{$_} or die "No $_ specified for use with MCWS\nStopped"; 
    }

    $self->{'urlpath'} = '/MCWS/v1';
    $self->{'url'} = URI->new('http://' . $self->{'server'});
    $self->{'ua'}->agent($self->{'agent'});
    $self->{'ua'}->timeout(10);
    $self->{'ua'}->env_proxy;
    $self->{'ua'}->cookie_jar({});
    $self->{'ua'}->credentials($self->{'server'}, $self->{'realm'}, $self->{'username'} => $self->{'password'});

    bless($self,$class);
    return $self;
}

# Send a list of items to MCWS
sub send {
    my $self = shift;
    my $list = shift;
    return unless $list;

    $self->send_items(\$_) for @$list;
}

sub send_items {
    my ($self, $cellsref) = @_;

    my %params;
    if (my @found = grep /filekey/i, keys $$cellsref) {
	$params{'FileType'} = 'Key';
	$params{'File'} = $$cellsref->{$found[0]};
    }
    else {
	$params{'FileType'} = 'Filename';
	$params{'File'} = uri_escape($$cellsref->{'Filename'}),
    }

    my (@fields, @values);
    debug "Filename: ", $params{'File'};
    while (my ($key, $value) = each %$$cellsref) {
	next if $key eq 'Filename' and $params{'FileType'} eq 'Filename';		# allow Filename changes when using FileKey.
	next if !defined $value or $value eq '' or $value eq '-';
	push @fields, $key;
	#push @values, $value;
	# XXX html entity decoding should not be done here, but for now, its easier.
	push @values, decode_entities($value);
	debug "Added field($key): ", $key, ' -> ', unfold_and_chop($value);
    }

    if (scalar @fields < 1) {
	verbose 'MCWS: Nothing to update for file: ', $params{'File'};
	return;
    }

    if (scalar @fields == 1) {
	$params{'Field'} = uri_escape_utf8($fields[0]);
	$params{'Value'} = uri_escape_utf8($values[0]);
    }
    else {
	$params{'List'}	= 'CSV';

	my $csv = Text::CSV->new ( { binary => 1, sep_char => ',' } );

	# Fields
	$csv->combine(@fields) or die "Failed to combine field names into CSV\nStopped";
	$params{'Field'} = uri_escape_utf8($csv->string());

	# Values
	$csv->combine(@values) or die "Failed to combine field values into CSV\nStopped";
	$params{'Value'} = uri_escape_utf8($csv->string());
    }

    $self->{'url'}->path(join '/', $self->{'urlpath'}, 'File/SetInfo');
    $self->{'url'}->query(join '&', map { join('=', $_, $params{$_}) } keys %params);
    debug "Sending: ", $self->{'url'}->canonical->as_string;

    my $response = $self->{'ua'}->get($self->{'url'}->canonical->as_string);
    unless ($response->is_success) {
	die "send_items: MCWS write failed: \n", $self->{'url'}->canonical->as_string, "\n", $response->status_line, "\n", $response->content;
    }
    $response->content =~ /<Response Status="OK"\/>/ or
	warn 'Unexpected server response: ', $response->content;
}

# Call MCWS to do an MCC command.
# Arguments are MCC command number and optional parameter
sub do_mcc_command {
    my $self = shift;

    #http://localhost:52199/MCWS/v1/Control/MCC?Command=10000
    #http://localhost:52199/MCWS/v1/Control/MCC?Command=10005&Parameter=2

    $self->{'url'}->path(join '/', $self->{'urlpath'}, 'Control/MCC');
    $self->{'url'}->query_form('Command', $_[0], scalar @_ > 1 ? ('Parameter', $_[1]) : '');
    debug "$_[0]: Sending: ", $self->{'url'}->canonical->as_string;

    my $response = $self->{'ua'}->get($self->{'url'}->canonical->as_string);
    unless ($response->is_success) {
	die "do_mcc_command: MCWS request failed: \n", $self->{'url'}->canonical->as_string, "\n", $response->status_line, "\n", $response->content;
    }
    $response->content =~ /<Response Status="OK"\/>/ or
	warn 'Unexpected server response: ', $response->content;

    # copying from the clipboard is a little slow with many items selected
    $_[0] == 21000 and sleep 1;
}

# Get the list of desired playlist types.
# @types is an array of playlist types to include: Group, Playlist, Smartlist.
sub get_playlists {
    my ($self, @types) = @_;
    my %playlists;

    # http://localhost:52199/MCWS/v1/Playlists/List
    $self->{'url'}->path(join '/', $self->{'urlpath'}, 'Playlists/List');

    my $response = $self->{'ua'}->get($self->{'url'}->canonical->as_string);
    unless ($response->is_success) {
	die "get_playlist: MCWS request failed: \n", $self->{'url'}->canonical->as_string, "\n", $response->status_line, "\n", $response->content;
    }

    $response->content =~ /<Response Status="OK">/ms or
	warn 'Unexpected server response: ', $response->content;

    local $_ = $response->content;
    while (/<Item>(.*?)<\/Item>/gms) {
	local $_ = $1;
	m#<Field\ Name="ID">([^<]+)</Field>\R
	  <Field\ Name="Name">([^<]+)</Field>\R
	  <Field\ Name="Path">([^<]+)</Field>\R
	  <Field\ Name="Type">([^<]+)</Field>
        #xms;
	next unless $4 ~~ @types;
	$playlists{$3} = { id => $1, name => $2, type => $4 };
    }
    return \%playlists;
}

# Get the list of files from the playlist specified by playlist ID.
sub get_playlist_contents {
    my ($self, $id) = @_;
    my @files;

    # http://localhost:52199/MCWS/v1/Playlist/Files?PlaylistType=ID&Action=mpl&Playlist=2013237201&Fields=Filename
    $self->{'url'}->path(join '/', $self->{'urlpath'}, 'Playlist/Files');
    $self->{'url'}->query_form('PlaylistType', 'ID', 'Action', 'MPL', 'Fields', 'Filename', 'Playlist', $id);

    my $response = $self->{'ua'}->get($self->{'url'}->canonical->as_string);
    unless ($response->is_success) {
	die "get_playlist_contents: MCWS request failed: \n", $self->{'url'}->canonical->as_string, "\n", $response->status_line, "\n", $response->content;
    }

    $response->content =~ /^<\?xml version="1\.0".+<\/MPL>\R$/ms or
	warn 'Unexpected server response: ', $response->content;

    local $_ = $response->content;
    while (/<Item>(.*?)<\/Item>/gms) {
	local $_ = $1;
	m#<Field\ Name="Key">[^<]+</Field>\R
	  <Field\ Name="Filename">([^<]+)</Field>\R
        #xms and push @files, $1;
    }
    return \@files;
}

1;
