package MCUtils::Misc 1.00;

use parent qw(Exporter);
our @EXPORT	= qw(get_clipboard_contents parse_clipboard debug verbose debug_on verbose_on unfold_and_chop);
our @EXPORT_OK	= qw(hexdump);

use diagnostics;
use v5.14;
use utf8;
use strict;
use warnings;

use Encode;

binmode STDOUT, ':encoding(UTF-8)';
binmode STDERR, ':encoding(UTF-8)';

my ($verbose, $debug);

sub debug {
    return unless $debug;

    printf "%-20s: %s\n", (split(/::/, (caller(1))[3] || 'main'))[-1], join('', @_);
}

sub verbose {
    return unless $verbose;

    say @_;
}

sub debug_on {
    $debug++;
}

sub verbose_on {
    $verbose++;
}

sub unfold_and_chop {
    local $_ = shift;

    return undef if not defined $_;
    s/\R/<CR>/g;
    my $len = length $_;
    return $_ ? (substr($_, 0, 77) . ($len > 77 ? '...' : '')) : '';
}

# Grab, decode and return the clipboard contents.
#
sub get_clipboard_contents {
    my $clip;

    if ($^O eq 'MSWin32') {
	$clip = Win32::Clipboard();
	$_ = $clip->GetAs(&Win32::Clipboard::CF_UNICODETEXT);
	$_ = Encode::decode("UTF16-LE", $_);
    }
    else {
	$clip = Mac::Pasteboard->new ();
	my ($data,$flags) = $clip->paste('public.utf8-plain-text');
	$_ = $data;
    }

    return $_;
}

sub parse_clipboard {
    local $_ = shift;
    my @cliptracks;

    # if there's a tab at the end of the line in the first row, we'll use the old format
    if (/^[^\n\r]+\t\R/) {
	my @clipheaders;

	# before MC 19.0.36: work around broken TSV format (no quoting was done)
	my ($headers,$rows) = split /\R+/, $_, 2;
	# bogus: MC outputs trailing \t on the header row
	# also strip any extra line feeds, newlines
	$headers =~ s/[\t\013\010]*$//;
	$headers = lc $headers; 		# canonicalize header names to lowercase.
	@clipheaders = split /\t/, $headers;

	my $n = @clipheaders - 1;
	while ($rows =~ s#^([^\t]*(?:\t[^\t]*){$n})\R##oms) {
	    my $r = $1;
	    $r =~ s/\s*\R$//;
	    debug "Row: '$r'\n";
	    my @row = split(/\t/, $r);
	    my %h = ();
	    for (my $i = 0; $i < @clipheaders; $i++) {
		$h{$clipheaders[$i]} = $row[$i];
	    }
	    push @cliptracks, \%h;
	}
    }
    else {
	# MC 19.0.36: use standard CSV format
	my $lref;
	my $sep_char = /\t/ ? "\t" : ',';
	my $csv = Text::CSV->new ( { binary => 1, sep_char => $sep_char } );
	$csv->eol("\015\012") if $^O eq 'darwin';
	open my $io, "<", \$_ or die $!;
	$lref = $csv->getline($io) or
	    die "Failed to get clipboard contents.\nStopped";
	$_ = lc foreach @$lref;		# canonicalize header names to lowercase.
	$csv->column_names ($lref);
	while (my $hr = $csv->getline_hr ($io)) {;
	    push @cliptracks, $hr;
	}
	close $io;
    }

    return \@cliptracks;
}

=cut
# Safe fork/exec to avoid cross-platform issues with qx// or backticks.  See perlsec.
# The open mode "-|" doesn't work with activeperl.
sub do_fork_exec {
    use English '-no_match_vars';
    my ($pid, $ret);;
    die "Can't fork: $!" unless defined($pid = open(KID, "-|"));
    if ($pid) {           # parent
	while (<KID>) {
	    $ret .= $_;
	}
	close KID;
	return $ret;
    }
    else {
	my @temp     = ($EUID, $EGID);
	my $orig_uid = $UID;
	my $orig_gid = $GID;
	$EUID = $UID;
	$EGID = $GID;
	# Drop privileges
	$UID  = $orig_uid;
	$GID  = $orig_gid;
	# Make sure privs are really gone
	($EUID, $EGID) = @temp;
	die "Can't drop privileges"
	    unless $UID == $EUID  && $GID eq $EGID;
	#$ENV{PATH} = "/bin:/usr/bin"; # Minimal PATH.
	# Consider sanitizing the environment even more.

	exec @_
	    or die "can't exec myprog: $!";
    }
}
=cut

# For a given string parameter, returns a string which shows
# whether the utf8 flag is enabled and a byte-by-byte view
# of the internal representation.
#
sub hexdump
{
    use Encode;
    my $str = shift;
    my $flag = Encode::is_utf8($str) ? 1 : 0;
    use bytes; # this tells unpack to deal with raw bytes
    my @internal_rep_bytes = unpack('C*', $str);
    return
        $flag
        . '('
        . join(' ', map { sprintf("%02x", $_) } @internal_rep_bytes)
        . ')';
}

1;
