# This scriplet uses the Echo Nest acoustic fingerprinting server to determine some song metadata such
# as artist, track name, various acoustic properties, and an MD5 signature which can be used for 
# duplicate detection.  The scriplet will return about twenty different values, each mapped to an
# MC field. You can customize the MC field names - see the Scriptlet Options section in the config file.
#
#	http://the.echonest.com/ 
#
# Requirements 
#   Windows:	need codegen executable in your path (32-bit version, tested in Win 7)
#		https://dl.dropboxusercontent.com/u/87189402/codegen.zip
#
#   Mac:	need echoprint-codegen executable in your path (64-bit tested on Mac OS X 10.9.1)
#		https://dl.dropboxusercontent.com/u/87189402/echoprint-codegen
#
#   All:	Need to install ffmpeg executable in your path:
#		http://www.ffmpeg.org/
#
#
# See also:
#   audio summary fields:	https://developer.echonest.com/raw_tutorials/faqs/faq_04.html
#   dancibility and energy:	http://runningwithdata.com/post/1321504427/danceability-and-energy
#   accousticness and valence:	http://developer.echonest.com/forums/thread/1297
#   duplicate detection:	http://yabb.jriver.com/interact/index.php?topic=87575.0


package AcousticFingerprint 1.05;

use v5.16;
use strict;
use warnings;

use MIME::Base64;
use JSON::XS;
#use Data::Dumper;

our %callbacks = (
    pre     => undef,
    perfile => \&do_perfile,
    post    => undef,
);

my $echonest_url = 'http://developer.echonest.com/api/v4/song/identify';
my $codegen_exe = $^O eq 'MSWin32' ? 'codegen' : 'echoprint-codegen';

my %options = (
    var => { 
	sample_start		=> 20,	# sample starting point (in seconds)
	sample_duration		=> 30,	# sample 30 seconds worth of audio
    },
    field => {
        # echo nest return keys to MC field names
        # key	    MC Field Name
	artist_name		=> 'Artist',
	title			=> 'Name',
	song_type		=> 'EN Song Type',		# Semicolon-separated list of: 'christmas', 'live' 'studio', 'acoustic', and 'electric'
	artist_discovery	=> 'EN Artist Discover',
	artist_id 		=> 'EN Artist ID',
	acousticness		=> 'EN Acousticness',
	audio_md5		=> 'EN MD5',
	danceability		=> 'EN Dancibility',		# range: 0 to 1
	energy			=> 'EN Energy',			# range: 0 to 1
	key			=> 'EN Key',		 	# key signature, starting at C and ascends the chromatic scale. 0 = C, 1 = D-flat, ...
	liveness		=> 'EN Liveness',
	loudness		=> 'EN Loudness',		# The overall loudness of a track in decibels (dB)
	mode			=> 'EN Mode',			# major or minor key: 0 = major, 1 = minor
	speechiness		=> 'EN Speechiness',
	tempo			=> 'EN Tempo',			# beats per minute
	time_signature		=> 'EN Time Signature',		# beats per measure
	valence			=> 'EN Valence',
	id			=> 'EN ID',
	message			=> 'EN Message',
	score			=> 'EN Score',
    },
);

$DB::single = 1;

main::set_scriptlet_opts(\%options);

my ($ua, $uri_params) = create_user_agent();

# Per-row callback function
sub do_perfile {
    my ($argstr, $field, $row) = @_;
    my ($json, $code, @names_to_try);
    my ($fileabbr) = $row->{'filename'} =~ m#((?:[\\/][^\\/]+){3})$#;

    my ($json, $code, @names_to_try);
    # Unicode confounds qx//, so try the ANSI name first, and if that fails, try the original name.  Some
    # files will fail, and fixing this is non-trivial since it requires setting up a different process
    # creation and STDOUT capture mechanism.
    if ($^O eq 'MSWin32') {
	use if $^O eq 'MSWin32', 'Win32';

	if (my $ansi_name = Win32::GetANSIPathName($row->{'filename'})) {
	    push @names_to_try, $ansi_name;
	}
	else {
	    warn "Win32::GetANSIPathName unable to get ANSI path name: $^E.\nStopped";
	}
    }

    my $i = 0;
    for (@names_to_try, $row->{'filename'}) {
	main::debug "\tTrying filename(", ++$i, '): ', $_;
	$json = qx($codegen_exe "$_" $options{'var'}{'sample_start'} $options{'var'}{'sample_duration'});
	if ($json =~ /"error":"could not decode"/) {
	    say 'codegen failed to decode for file($i): ', $_;
	    next;
	}
	($code) = ($json =~ /"code":"([^"]+)"/);
	last if $code;
    }
    return undef unless $code;

    # there should be good codes at this point
    my $file_url = join '?', $echonest_url, join('&', $uri_params, join('=', 'code', $code));

    #main::debug "Sending: $file_url";
    my $response = $ua->get($file_url);
    $response->is_success or
	die "HTTP get failed: \n$file_url\n", $response->status_line, "\n", $response->content;

    main::debug "\t\tRESP: ", $response->content;
    my $r = (decode_json($response->content))->{'response'};

    if ($r->{'status'}{'message'} ne 'Success') {
	say "\tStatus = ", $r->{'status'}{'message'}, '; ...', $fileabbr;
	return undef;
    }
    my %ret;
    if (! $r->{'songs'}[0]) {
	say "\tTrack lookup - No results found: ...", $fileabbr;
	return undef;
    }

    my $rr = $r->{'songs'}[0];
    for my $k1 (keys $rr) {
	if (ref $rr->{$k1} eq 'HASH') {
	    for my $k2 (keys $rr->{$k1}) {
		next if $k2 eq 'analysis_url';
		next if $k2 eq 'duration';
		$ret{$options{'field'}{$k2}} = $rr->{$k1}{$k2} if defined $rr->{$k1}{$k2};
		main::debug 'Key/Value: ', $k2, ' = ', $rr->{$k1}{$k2} || 'undef';
	    }
	}
	else {
	    my $v;
	    if (ref $rr->{$k1} eq 'ARRAY') {
		$v = join '; ', @{$rr->{$k1}};
	    }
	    else {
		$v = $rr->{$k1};
	    }
	    main::debug 'Key/Value: ', $k1, ': ', $v || 'undef';
	    $ret{$options{'field'}{$k1}} = $v   if defined $v;;
	}
    }

    return \%ret;
}

sub create_user_agent {
    chomp(my $api_key = decode_base64('VjlUU1hZR0lYVVBYRFhIUVoK'));
    my @params = (
	[ 'version' => '4.12'],
	[ 'bucket'  => 'audio_summary'],
	[ 'bucket'  => 'artist_discovery'],
	[ 'bucket'  => 'song_type'],
	[ 'api_key' => $api_key],
    );

    my $ua = LWP::UserAgent->new(agent => 'pscriptor/AcousticFingerprint');
    $ua->timeout(10);
    $ua->env_proxy;
    $ua->cookie_jar({});

    return ($ua, join '&', map { join '=', $_->[0], $_->[1] } @params);
}

1;
