=cut
This scriplet will copy selected files and create M3U8 playlist files for those files.  The
destination folders and file names are highly customizable.  Destination folder names are
constructed in three parts:

	<base folder> + <sub-folder fragment> + <filename>

You will specify the <base folder>.  The <sub-folder fragment> is constructed from:

	<playlist group> + <expression column>

Either of these components may be empty, where no <sub-folder fragment> would be added. The
<playlist group> gives you the ability to create multple playlists + file copies, each in their
own folder, and the <expression column> allows you to place tracks for the given playlist under
sub-folders, such as [Artist]\[Album].  You specify the column containing the <playlist group>
and the <expression column>'s name (the column does not have to be an <expression column> - any
column can be used).

Each playlist file is constructed within a <playlist group> folder, and it contains relative
path names, so that you can drag a <playlist group> folder elsewhere.

Each <filename> is constructed from a template you may provide (which will default to using the
file's existing filename).  The template is constructed using "tokens", one of them being a
sequencing index, allowing you to add sequence numbers to the files.  Other tokens allow using
values from the view columns and the file's current name.

Example usage:

    perl pscriptor.pl -c pscriptor-config.txt -E CopyWithM3U8 -f 'MyGroupField' -- --base "M:\My Copied Music" --folder Playlists --filename "<seq> <fname>"

Requirements
    Windows:
	ppm install Win32-Unicode

See Also:
    http://yabb.jriver.com/interact/index.php?topic=85990.msg611027#msg611027
    http://yabb.jriver.com/interact/index.php?topic=85990.msg611361#msg611361
=cut

package CopyWithM3U8 1.00;

use v5.16;
use strict;
use warnings;
use Encode;
use Encode::Locale;
use Getopt::Long;
use File::Basename;
use File::Spec;
use File::Copy;
use if $^O eq 'MSWin32', 'Win32::Unicode::File';
use if $^O eq 'MSWin32', 'Win32::Unicode::Dir';

our %callbacks = (
    pre     => \&do_preprocessing,
    perfile => undef,
    post    => undef,
);

$DB::single = 1;

my %tracks_per_group;
my ($dst_basepath, $dst_folderfield, $dst_namespec);
my $playlist_name = 'Playlist.m3u8';

my $package = __PACKAGE__;

sub Usage {
    my $exitcode = shift;
    say @_ ? join('', @_, "\n") : '',
    <<ENDUSAGE, "\nStopped";

$package options:
    --basepath    | -b <base path>	# base path of the destination folder
    --delete      | -d			# delete existing files first?
    --help	  | -h			# output scriplet help and usage text
    --filename    | -f <filename spec>	# the specification that defines how filenames should be constructed
    --folderfield | -F <column name>	# the name of a column that defines the sub-folder component
ENDUSAGE
    exit $exitcode;
}

sub do_preprocessing {
    my ($argstr, $tracksref, $fieldsref) = @_;

    @$fieldsref == 1 or die "Only a single -f field is permitted with ", $package;

    my $delete_ok;
    my $groupfield = $fieldsref->[0];

    my %opts = ();
    {
    local $SIG{__WARN__} = sub { say "\n*** ", $_[0]; };

    main::debug "Scriptlet args: @ARGV";
    Getopt::Long::Configure('no_ignore_case', );
    GetOptions(\%opts,
	    'basepath|b=s',
	    'delete',
	    'help|h'          => sub { Usage(0) },
	    'filename|f=s',
	    'folderfield|F=s' => sub { $opts{$_[0]} = lc $_[1] },
	) or Usage(1);
    }

    @ARGV == 0 or Usage 1, "Unexpected scriptlet arguments remaining: ", join " ", @ARGV;
    $dst_basepath = $opts{'basepath'} // Usage 1, "Required --base option is missing";
    $dst_basepath ne clean_path($dst_basepath, 0) and
	die $package, ": Destination base path is not a clean base path - it contains filesystem illegal characters.\nStopped";

    $dst_folderfield = $opts{'folderfield'} // undef;
    if ($dst_folderfield and not exists $tracksref->[0]->{$dst_folderfield}) {
	warn $package, ": folderfield was specified, but does not exist in the view.\n";
    }
    $dst_namespec = $opts{'filename'} // '<fname>';
    $delete_ok = exists $opts{'delete'};

    for (@$tracksref) {
	push @{$tracks_per_group{$_->{$groupfield}}}, $_;
    }

    # run through the tracks_per_group for the list of files to be included in the playlist
    foreach my $group (sort keys %tracks_per_group) {

	my $dst_group = clean_path($group, 1);
	my $playlist_dir  = File::Spec->catfile($dst_basepath, $dst_group);
	my $playlist_path = File::Spec->catfile($playlist_dir, $playlist_name);
	main::verbose "\t* creating playlist's folder path: ", $playlist_dir;
	next unless ! mkdir_recursive($playlist_dir);

	main::verbose "\t** make playlist file: ", $playlist_path;
	open my $plfile, $^O eq 'MSWin32' ? ">:encoding(UTF-8)" : '>', encode(locale_fs => $playlist_path) or
	    die "Can't create playlist file $playlist_path\n$!\nStopped";

	my $i = 0;
	my (%src_tracks, %copied_tracks);
	foreach my $t (@{$tracks_per_group{$group}}) {
	    $i++;

	    main::verbose "\ttrack: ", $t->{'filename'};

	    my ($f,$p,$s) = fileparse($t->{'filename'}, qr/\.[^.]+/);

	    my $dst_filename = $dst_namespec;
	    $dst_filename =~ s/<seq(?:,(\d+))?>/sprintf '%0' . ($1 || '') . 'd', $i/e;
	    $dst_filename =~ s/<fname>/$f/e;
	    while ($dst_filename =~ s!<\[([^]]+)\]>!$t->{lc $1} // ''!ge) {
		main::verbose "\t\ttoken replace: ", $1, ' => ', $t->{$1} // '';
	    }
	    $dst_filename .= $s;

	    my $dst_folderpath = $dst_folderfield ? clean_path($t->{$dst_folderfield} // '', 0) : '';
	    $dst_folderpath =~ s/^[\\\/]+//;
	    my $dst_path = File::Spec->catfile($dst_basepath, $dst_group, $dst_folderpath);
	    my $dst_filepath = File::Spec->catfile($dst_path, $dst_filename);

	    my $skip_copy = 0;
	    if (exists $src_tracks{$t->{'filename'}}) {
		main::verbose "\t! dup track in playlist: ", $t->{'filename'};
		if ($src_tracks{$t->{'filename'}} eq $dst_filepath) {
		    main::verbose "\t!! copy skipped (copied previously): ", $dst_filepath;
		    $skip_copy = 1;
		}
	    }

	    main::verbose "\t+ creating track's folder path: ", $dst_path;
	    next unless ! mkdir_recursive($dst_path);

	    my $actual_dst_filepath = $dst_filepath;
	    if (!$skip_copy) {
		if ($delete_ok and not $dst_filepath ~~ %copied_tracks) {
		    main::verbose "\t-- deleting existing track: ", $dst_filepath;
		    unless (do_unlink($dst_filepath)) {
			warn "Skipping file: failed to delete existing file: ", $dst_filepath, ': ', $!, "\n";
			next;
		    }
		}
		if ($actual_dst_filepath = do_copy($t->{'filename'}, $dst_filepath)) {
		    # remember that this src track has been copied, and remember its dst name
		    $src_tracks{$t->{'filename'}} = $actual_dst_filepath;
		    # remember the copied tracks, so we can delete/overwrite when the option is enabled
		    $copied_tracks{$actual_dst_filepath}++;
		}
	    }

	    main::verbose "\t+++ adding track to playlist: $actual_dst_filepath";
	    my $entry = File::Spec->catfile('.', $dst_folderpath, basename($actual_dst_filepath));
	    $entry = '.\\' . $entry if $^O eq 'MSWin32';	# catfile removes leading '.\' on Win32
	    say $plfile $entry;
	}
	close $plfile or
	    die "Failed to close playlist file $playlist_path\n$!\nStopped";
    }

    return undef;
}

sub mkdir_recursive {
    my $upath = shift;

    my $sep = $^O eq 'MSWin32' ? '\\' : '/';
    my @dirs = split /\Q$sep\E/, $upath;
    my $d = shift @dirs;
    for (@dirs) {
	$d = join $sep, $d, $_;
	unless (check_exists($d)) {
	    next unless ! ($^O eq 'MSWin32' ? mkdirW $d : mkdir $d);
	    warn "Failed to make directory ", $d, ": $!\n";
	    return 1
	}
    }
}

sub do_copy {
    my ($src, $dst) = @_;

    my $i = 0;
    while (check_exists($dst)) {
	$dst =~ s/ \((\d+)\)$//;
	$i = 1 + ($1 ||  $i);
	$dst .= " ($i)";
	main::verbose "\t\tTrying new name: ", $dst;
    }

    my $copy_ret;
    main::verbose "\t++ copying track to destination: ", $src, ' ==> ', $dst;
    if ($^O eq 'MSWin32') {
	$copy_ret = copyW($src, $dst);
    }
    else {
	$copy_ret = File::Copy::copy($src, $dst);
    }
    return $dst if $copy_ret;

    warn "! Copy Failed: $src ==> $dst\n$!\n";
    return undef;
}

sub check_exists {
    local $_ = shift;
    return $^O eq 'MSWin32' ? file_type('e', $_) : -e $_;
}

sub do_unlink {
    local $_ = shift;
    return 1 if ! check_exists($_);
    return $^O eq 'MSWin32' ? unlinkW($_) : unlink $_;
}

# 
sub clean_path {
    (local $_, my $remove_sep) = @_;
    my $drive;
    my $illegalchars = ':*?"<>|';
    if ($^O eq 'MSWin32') {
	$illegalchars .= $remove_sep ? '\\/' : '/';
	s/^([A-Z]:)//;
	$drive = $1;
    }
    else {
	$illegalchars .= $remove_sep ? '\\/' : '\\';
	s!(/[.]+)!/_!g;
    }
    my $re = qr/[\Q$illegalchars\E]/;
    s/$re/_/g;
    ($drive || '') . $_;
}

1;
