# Attempts to find and unify dates found in the supplied field(s).
#
# Requirements
#
#   Windows:    Needs the Date::Manip module.  Install via Windows command shell by
#               entering the command:
#
#	           ppm install Date-Manip 
#

package DateParse 1.05;

use v5.16;
use strict;
use warnings;

use Date::Manip::Date;

our %callbacks = (
    pre     => undef,
    perfile => \&do_perfile,
    post    => undef,
);

my %options = (
    var => {
	date_separator => '(?<sep>[-.])',
	date_output_template => '%Y-%m-%d',
    },
);

$DB::single = 1;

main::set_scriptlet_opts(\%options);

# These Date::Manip::parse_format() date templates are tried sequentially until a match is found
# or failure.
#
#   http://search.cpan.org/~sbeck/Date-Manip-6.43/lib/Date/Manip/Date.pod#parse_format
#

# Templates should be ordered most-to-least restrictive.
# feel free to comment out any that you don't need, or that might cause misdetection.
# Add a # character at the front of any entry to comment it out.
my @date_formats = (
    # strict, common formats first
    '(?:^|\D)%Y' . $options{'var'}{'date_separator'} . '%m\k<sep>%d(?:\D|$)',	# YYYY-MM-DD
    '(?:^|\D)%m' . $options{'var'}{'date_separator'} . '%d\k<sep>%Y(?:\D|$)',	# MM-DD-YYYY
    '(?:^|\D)%d' . $options{'var'}{'date_separator'} . '%m\k<sep>%Y(?:\D|$)',	# DD-MM-YYYY

    '(?:^|\D)%m' . $options{'var'}{'date_separator'} . '%e\k<sep>%Y(?:\D|$)',	# MM-D-YYYY
    '(?:^|\D)%f' . $options{'var'}{'date_separator'} . '%d\k<sep>%Y(?:\D|$)',	# M-DD-YYYY
    '(?:^|\D)%d' . $options{'var'}{'date_separator'} . '%f\k<sep>%Y(?:\D|$)',	# DD-M-YYYY
    '(?:^|\D)%e' . $options{'var'}{'date_separator'} . '%f\k<sep>%Y(?:\D|$)',	# D-MM-YYYY


    '(?:^|\D)%Y' . $options{'var'}{'date_separator'} . '%f\k<sep>%d(?:\D|$)',	# YYYY-M-DD
    '(?:^|\D)%Y' . $options{'var'}{'date_separator'} . '%m\k<sep>%e(?:\D|$)',	# YYYY-MM-D

    '(?:^|\D)%f' . $options{'var'}{'date_separator'} . '%e\k<sep>%Y(?:\D|$)',	# M-D-YYYY
    '(?:^|\D)%Y' . $options{'var'}{'date_separator'} . '%f\k<sep>%e(?:\D|$)',	# YYYY-M-D

    # double-digit year abbreviated forms next 
    '(?:^|\D)%d' . $options{'var'}{'date_separator'} . '%m\k<sep>%y(?:\D|$)',	# DD-MM-YY
    '(?:^|\D)%m' . $options{'var'}{'date_separator'} . '%d\k<sep>%y(?:\D|$)',	# MM-DD-YY

    # single-digit month or days
    '(?:^|\D)%d' . $options{'var'}{'date_separator'} . '%f\k<sep>%y(?:\D|$)',	# DD-M-YY
    '(?:^|\D)%f' . $options{'var'}{'date_separator'} . '%d\k<sep>%y(?:\D|$)',	# M-DD-YY
    '(?:^|\D)%f' . $options{'var'}{'date_separator'} . '%m\k<sep>%y(?:\D|$)', # D-MM-YY
    '(?:^|\D)%f' . $options{'var'}{'date_separator'} . '%e\k<sep>%y(?:\D|$)',	# M-D-YY

    # year-only or year,month-only isn't supported by Date::Parse

);

my $date  = new Date::Manip::Date;
my $today = $date->new_date;
$today->parse('today');

# Per-row callback function
sub do_perfile {
    my ($argstr, $field, $row) = @_;

    for (@date_formats) {
	my $err = $date->parse_format($_, $row->{$field}, qw/noiso8601 nodow nocommon noother nospecial nodelta noholidays/);
	if (! $err) {
	    if ($date->cmp($today) > 0) {
		main::verbose "Rejecting date later than today ", $date->printf($options{'var'}{'date_output_template'}), ": ", $_;
		next;
	    }
	    main::verbose $date->printf("\tDateParse: date found: year: %Y, month: %m, day: %d"), ': format: ', $_;
	    return $date->printf($options{'var'}{'date_output_template'});
	}
	if ($err ne '1') {
	    say "DateParse: parse_format error: ", $err;
	    next;
	}

	main::debug "\tDateParse: No match on format: ", $_;
    }

    # try some incomplete dates
    # try month-year, defaulting to 1 as the day.
    return join('-', $2, $1, '1')	if $row->{$field} =~ /(0?[1-9]|1[012])-(19\d{2}|20[01]\d)\b/;
    # Year-only values now...
    return $1	if $row->{$field} =~ /\b(19\d{2}|20[01]\d)\b/;

    return undef;
}

1;
