# Obtains cover art size from embedded cover art, and by default returns dimensions as WxH.
# If you want dimensions in two separate fields, use -a fieldW,fieldH where the
# fields fieldW and fieldH are your width and height field names, respectively.  You'll still
# need to include a -f, however, it will not be used, so just use the name of any field in the view.
#
# Requirements
#    Image::ExifTool.
#
#   Windows:   Install with the command line:  ppm install Image-ExifTool 

package GetCoverArtSize 1.05;

use v5.16;
use strict;
use warnings;

$DB::single = 1;

our %callbacks = (
    pre     => \&do_preprocessing,
    perfile => \&do_perfile,
    post    => undef,
);

use Image::ExifTool;

my @fields;
my $exifTool = new Image::ExifTool;

# Pre-processing callback function
sub do_preprocessing {
    my ($argstr, $tracksref, $fieldsref) = @_;

    if ($argstr) {
	@fields = split /\s*,\s*/, lc $argstr;
	scalar @fields == 2 or
	    die "When using -a, ", __PACKAGE__, " requires the names of two comma-separated output fields.\nExample: -a fieldW,fieldH\nStopped";
    }

    return undef;
}

# Per-row callback function
sub do_perfile {
    my ($argstr, $field, $row) = @_;

    # Unicode confounds qx//, so try the ANSI name first, and if that fails, try the original name.  Some
    # files will fail, and fixing this is non-trivial since it requires setting up a different process
    # creation and STDOUT capture mechanism.
    my (@names_to_try);
    if ($^O eq 'MSWin32') {
	use if $^O eq 'MSWin32', 'Win32';

	if (my $ansi_name = Win32::GetANSIPathName($row->{'filename'})) {
	    push @names_to_try, $ansi_name;
	}
	else {
	    warn "Win32::GetANSIPathName unable to get ANSI path name: $^E.\nStopped";
	}
    }

    my $info;
    my $i = 0;
    for (@names_to_try, $row->{'filename'}) {
	main::debug "\tTrying filename(", ++$i, '): ', $_;
	$info = $exifTool->ImageInfo($_);
	last if $info;

	main::debug "Image::ExifTool::ImageInfo failed to open file($i): ", $_;
    }
    return undef unless $info;

    if (exists $info->{'Picture'}) {
	my $embedinfo = $exifTool->ImageInfo($info->{'Picture'});
	if (exists $embedinfo->{'ImageSize'}) {
	    #print "Embedded image: size $embedinfo->{'ImageSize'}\n";
	    if (@fields) {
		my ($width,$height) = split /x/, $embedinfo->{'ImageSize'};
		return { $fields[0] => $width, $fields[1] => $height };
	    }
	    else {
		return $embedinfo->{'ImageSize'};
	    }
	}
    }

    return undef;
}

1;
