# MakeBDChapters
#
# Creates particle-based Blu-ray chapters from the selected Blu-Ray MPLS file.
#
# Requirements:
#
# You'll need to do the following:
#
#     - get the mediainfo command line tool: http://mediaarea.net/en/MediaInfo/Download
#     - customize a view's file list, as per the instructions in the Particlize scriptlet
#     - configure the MakeBDChapters section in your custom config
#
# This scriptlet works in two phases:
#
# Phase 1: an mpls playlist file in MC is selected, and MakeBDChapters will call
#          mediainfo to collect the chapter information, create and import dummy files.
# Phase 2: the mpls file and the dummy files are selected, and MakeBDChapters will
#          populate tags for the dummy files, and call Particlize to connect the mpls
#	   with the new dummy file entries (which then deletes the dummy files).
#
# For Phase 1, perform the following steps:
#   1. Select your blu-ray entry in MC (it will have a Filename (name) of "index.bluray;1".
#   2. Right-click the entry and select Stacks > Advanced > Auto Create BD/DVD Title Particles
#   3. Select the desired MPLS file that was created.
#   4. Run pscriptor to create and import the necessary dummy files.  These will be used and
#      removed in phase 2.
#
# For Phase 2, perform the following steps:
#   1. Select the same mpls file and the imported dummy files.
#   2. Run the same pscriptor command used for phase 1.
#
# An example command line is:
#
#     perl pscriptor.pl -E MakeBDChapters -f "tmp field" --filekey
#
# The field 'tmp field' is used in phase 1 to store the chapter playback range information
# in the mpls file for later retrieval in phase 2.  You can use any field you like, and
# the field must be in the file list.
#
# You must supply the --filekey argument to pscriptor, or the partilization will fail.
#
# See also this thread and subsequent replies:
#
#    http://yabb.jriver.com/interact/index.php?topic=85990.msg613942#msg613942

package MakeBDChapters 1.00;

use v5.16;
use strict;
use warnings;

use File::Spec;
use MCUtils::DummyFile;

our %callbacks = (
    pre     => \&do_preprocessing,
    perfile => \&do_perfile,
    post    => \&do_postprocessing,
);

my %options = (
    var => { 
	exe_path_win	=> 'mediainfo.exe',
	exe_path_mac	=> 'mediainfo',
	import_cmd	=> 'mc19.exe /Import',             # command to import dummy files
	basedir		=> '$DESKTOP',	# default basedir path (note uninterpolated $DESKTOP macro)
    },
);

$DB::single = 1;

main::set_scriptlet_opts(\%options);

my ($exe, $desktop, $dummyfile_dir, @ranges);
my $curfile_index = 0;

if ($^O eq 'MSWin32') {
    $exe = $options{'var'}{'exe_path_win'};
    $desktop = File::Spec->catdir($ENV{'USERPROFILE'}, 'Desktop');
}
else {
    $exe = $options{'var'}{'exe_path_mac'};
    $desktop = File::Spec->catdir($ENV{'HOME'}, 'Desktop');
}

# expand $DESKTOP macro to path to Desktop
$options{'var'}{'basedir'} =~ s/^\$DESKTOP[\\\/]*/${desktop}/;

my $phase = 1;
sub do_preprocessing {
    my ($keyfield, $tracksref, $fieldsref) = @_;

    -e $exe or die "Command '$exe' cannot be found\nStopped";
    -x $exe or die "command '$exe' is not executable.\nStopped";

    # when only a single track is present, assume its a playlist, so this is phase 1.
    return undef if @$tracksref == 1;

    # sort the tracksref to place the main file at the front of the list
    # the top file (index 0) will be the main file; the dummy files 1..n will follow.
    my @tracks = sort by_filename @$tracksref;
    $tracksref = \@tracks;

    # Particlize requires the Album name, so use the top file's Album name.
    for (my $i = 1; $i < scalar @$tracksref; $i++) {
	$tracksref->[$i]->{'album'} = $tracksref->[0]->{'album'};
    }

    require Particlize;
    $tracksref = Particlize::do_preprocessing(@_);

    $phase = 2;

    return $tracksref;
}

# Per-row callback function
sub do_perfile {
    my ($argstr, $field, $row) = @_;

    my $info;
    # Phase 1: get the media info, and stuff the list of playback ranges into the mpls' Playback Range
    # for later parsing
    if ($phase == 1) {
	if ($row->{'filename'} =~ /(\d+)\.mpls/) {
	    my $playlist_id = $1;
	    $dummyfile_dir = File::Spec->catdir($options{'var'}{'basedir'}, 'MakeBDChapters', $row->{'album'});
	    $info =  get_mediainfo($row->{'filename'});
	    for (my $i = 0; $i < scalar @{$info->{'chapters'}}; $i++) {
		create_dummy_file($dummyfile_dir, exists $main::opts{'testonly'}, $playlist_id, $i + 1);
	    }
	}
	else {
	    (my $playlist_dir = $row->{'filename'}) =~ s/index.bluray;1/PLAYLIST/;
	    opendir(my $dh, $playlist_dir) || die "Failed to open BD's PLAYLIST directory: $!\n$row->{'filename'}\nStopped";
	    while(readdir $dh) {
		next if not /\.mpls$/;
		$info = get_mediainfo("$playlist_dir\\$_");
	    }
	    closedir $dh;
	}

	return {
	    chapters	=> scalar @{$info->{'chapters'}},
	    $field 	=> $info->{'ranges'},
	};
    }

    # Phase 2: set the playback ranges for each dummy file, and particlize each file entry
    my %h;

    if ($curfile_index == 0) {
	@ranges = split /; /,  $h{$field} = $row->{$field};
    }
    else {
	$h{'Playback Range'} = $ranges[$curfile_index - 1];
	$h{'Track #'}        = $curfile_index;
	main::debug "    Chapter", $curfile_index;
    }

    $curfile_index++;
    my $ret = Particlize::do_perfile(@_);
    return { %h, %$ret };
}

sub do_postprocessing {
    if ($phase == 1) {
	if ($options{'var'}{'import_cmd'}) {
	    my @cmd = split / +/, $options{'var'}{'import_cmd'};
	    system @cmd, "$dummyfile_dir";
	}
    }
    else {
	Particlize::do_postprocessing();
    }
}

sub get_mediainfo {
    my %info = (
	playlist => shift,
    );

    my $results = qx($exe, "--full", "$info{'playlist'}");
    ($info{'duration'}) = ($results =~ /^Duration\s+: (\d+:\d+:\d+\.\d+)$/gm);

    main::debug 'Playlist: ', $info{'playlist'};
    main::debug 'Duration: ', $info{'duration'};

    while ($results =~ /^(\S+)\s+: Chapter (\d+)$/gm) {
	main::debug "\tChapter $2: $1";
	push @{$info{'chapters'}}, $1;
    }
    for (@{$info{'chapters'}}) {
	$info{'ranges'} .= $info{'ranges'} ? '-' . $_ . '; ' . $_ : $_;
    }
    $info{'ranges'} .= '-' . $info{'duration'};

    return \%info;
}

# sorts trackref array by filename, main file up top, followed by dummy files.
sub by_filename {
    return -1 if $a->{'filename'} !~ /\d+\.mp3$/;
    return  1 if $b->{'filename'} !~ /\d+\.mp3$/;
    my ($aname) = ($a->{'filename'} =~ /(\d+)\.mp3$/);
    my ($bname) = ($b->{'filename'} =~ /(\d+)\.mp3$/);
    return $aname <=> $bname;
}

1;
