# Makes a random m3u8 playlist file consisting of N tracks from your selected tracks.  The file will
# by default be named RandomPlaylist.m3u8 and will be placed on the Desktop, or you can specify
# the full path to a filename via --args.  In addition you can specify the quantity of files to be
# added to the playlist, a playback zone, and whether or not playback should autostart.  The --args
# format is:
#
#    --args "n=2 zone=2 path=<fullpath> play=1"
#
# where any of the fields may be ommited (whereby config file or default values will be used).
# See the MakeRandomPlaylist section under Scriptlet Options in the config file.  Example usage:
#
#    perl pscriptor ... -E MakeRandomPlaylist --args "n=25 path=C:\Users\MrC\Desktop\playlist.m3u8"
#
# Note: Be sure to place double quotes around the --args argument, as shonw in the example above.
#
# See: http://yabb.jriver.com/interact/index.php?topic=85990.msg606255#msg606255

package MakeRandomPlaylist 1.00;

use v5.16;
use strict;
use warnings;

use Encode;

our %callbacks = (
    pre     => \&do_preprocessing,
    perfile => undef,
    post    => \&do_postprocessing,
);

my %options = (
    var => { 
	zone		=> -1,					# default playback zone (-1 means default zone)
	num_tracks	=> 25,					# default number of tracks to include in the playlist
	autoplay	=> 1,					# autostart playback
	autoplay_cmd	=> 'mc19.exe /PlayReplace',		# autostart playback command
	playlist_path	=> '$DESKTOP/RandomPlaylist.m3u8',	# default playlist path (note uninterpolated $DESKTOP macro)
    },
);

my ($pathsep, $desktop, $fileh, @tracks);

$DB::single = 1;

if ($^O eq 'MSWin32') {
    $desktop = join($pathsep = '\\', $ENV{'USERPROFILE'}, 'Desktop');
}
else {
    $desktop = join($pathsep = '/', $ENV{'HOME'}, 'Desktop');
}

main::set_scriptlet_opts(\%options);

sub do_preprocessing {
    my ($argstr, $tracksref, $fieldsref) = @_;

    # -a 'n=2 zone=2 path=field'
    if ($argstr) {
	($argstr =~ s/\bn=([\d]+)\b//g)		and $options{'var'}{'num_tracks'} = $1;
	($argstr =~ s/\bzone=([\d]+)\b//g)	and $options{'var'}{'zone'} = $1;
	($argstr =~ s/\bplay=([\d])\b//g)	and $options{'var'}{'autoplay'} = $1;

	if ($argstr =~ s/\bpath=(.+)$//) {
	    $options{'var'}{'playlist_path'} = $1;
	    $options{'var'}{'playlist_path'} =~ s/\s*$//;
	}

	$argstr =~ s/\s//g;
	$argstr ne '' and die "Unknown --args argument '$argstr' supplied to ", __PACKAGE__;
    }

    # expand $DESKTOP macro to actual path to Desktop
    $options{'var'}{'playlist_path'} =~ s/^\$DESKTOP[\\\/]+/${desktop}${pathsep}/;

    main::verbose 'make playlist file: ', $options{'var'}{'playlist_path'};
    open $fileh, ">:encoding(UTF-8)", encode(locale_fs => $options{'var'}{'playlist_path'}) or
	die "Can't create playlist file $options{'var'}{'playlist_path'}\n$!\nStopped";

    for (@$tracksref) {
	push @tracks, $_->{'filename'};
    }

    return undef;
}

sub do_postprocessing {
    my $ntracks = scalar @tracks;

    for (my $i = 0; $i < $ntracks and $i < $options{'var'}{'num_tracks'}; $i++) {
	# select N random tracks
	my $rnd = int(rand $ntracks);
	main::verbose "\tAdded track: ", $tracks[$rnd];
	say $fileh $tracks[$rnd];
    }

    close $fileh or
	die "Failed to close playlist file $options{'var'}{'playlist_path'}\n$!\nStopped";

    if ($options{'var'}{'autoplay'}) {
	my @cmd = split / +/, $options{'var'}{'autoplay_cmd'};
	push @cmd, join '', $options{'var'}{'playlist_path'}, 
			    $options{'var'}{'zone'} eq '-1' ? () : ('|Zone=', $options{'var'}{'zone'});

	system @cmd;
    }

    return;
};

1;
