# Returns a sequence of numbers defined by the arguments passed via -a.  A starting number
# and possible increment can be supplied (both default to 1).  The counter restarts at the
# beginning of each group, if defined.  The format of the -a argument is as follows:
#
#   -a "seq=<start number>+<increment> group=<grouping field name>"
#
# where <start number> is the starting number of the sequence, +<increment> is the increment
# to be added to the previous value (the +<increment> is optional, and defaults to 1).  The
# <grouping field name> defines the grouping of files, where withing each group, the sequence
# of numbers restarts at the <start number>. If the field=<grouping field name> is absent,
# the grouping will be the entire selection.
#
# Note: Be sure to include quotes around the -a argument string if it contains any whitespace.

package Numberize 1.01;

use v5.16;
use strict;
use warnings;

our %callbacks = (
    pre     => \&do_preprocessing,
    perfile => \&do_perfile,
    post    => undef,
);

my $groupfield_default = '_+_ def group field _+_';
my (%groups, $groupfield);
my ($start, $inc) = (1, 1);

$DB::single = 1;

sub do_preprocessing {
    my ($argstr, $tracksref, $fieldsref) = @_;

    # -a 'seq=2+2 group=field'
    if ($argstr) {
	if ($argstr =~ s/seq=(?<start>-?[.\d]+)(?:\+(?<inc>-?[.\d]+))?//g) {
	    $start = $+{start};
	    $inc = $+{inc} if exists $+{inc};
	}
	if ($argstr =~ s/\s*group=(.+)$//) {
	    $groupfield = lc $1;
	    $groupfield =~ s/\s*$//;
	    $groupfield ~~ %{$tracksref->[0]} or die "The grouping field '$groupfield' supplied to -a does not exist in the view.\nStopped";
	}

	$argstr ne '' and die "Unknown -a argument '$argstr' supplied to ", __PACKAGE__;
    }

    return undef;
}

sub do_perfile {
    my ($argstr, $field, $row) = @_;

    my $gf = defined $groupfield ? $row->{$groupfield} : $groupfield_default;

    if (!exists $groups{$gf}) {
	$groups{$gf} = $start;
    }
    else {
	$groups{$gf} += $inc;
    }

    return $groups{$gf};
}

1;
