# Particlize
#
# Converts dummy files (like those created by discogs.pl or amg.pl) into particles.
# 
# This scriptlet makes particles from a selected file by using some imported dummy files
# which may have been created earlier by discogs.pl, amg.pl or another scriplet such as
# MakeBDChapters.
#
# Create a customized view with the following fields in file list:
#
#    - the Filename column
#    - an expression column named FileKey with the expression: FileKey()
#    - an expression column named Stack Top with the expression: [Stack Top]
#    - an expression column named Stack Files with the expression: [Stack Files]
#
# The view should be customized to show now only the desired main file, but also to
# include files from a dummy file directory you will specify when you run the command.
#
# You can use the Search box to selectively filter to a single title to be particlized.
#
# See this thread for a workflow example:
#
#    http://yabb.jriver.com/interact/index.php?topic=85990.msg613786#msg613786

package Particlize 1.00;

use v5.16;
use strict;
use warnings;

use File::Copy;

our %callbacks = (
    pre     => \&do_preprocessing,
    perfile => \&do_perfile,
    post    => \&do_postprocessing,
);

$DB::single = 1;

my (@filekeys, $top_filename, $nfiles, %pushup_fields, %pushdown_fields);

# sorts trackref array by filename, main file up top, followed by dummy files.
sub by_filename {
    return -1 if $a->{'filename'} !~ /\d+\.mp3$/;
    return  1 if $b->{'filename'} !~ /\d+\.mp3$/;
    my ($aname) = ($a->{'filename'} =~ /(\d+)\.mp3$/);
    my ($bname) = ($b->{'filename'} =~ /(\d+)\.mp3$/);
    return $aname <=> $bname;
}

sub do_preprocessing {
    my ($keyfield, $tracksref, $fieldsref) = @_;

    my @pushups   = ('Album');
    my @pushdowns = ('File Type', 'Media Type', 'Media Sub Type', 'Album', 'Artist');
    my @required_fields = ('Filename', 'Stack Top|[Stack Top]', 'Stack Files|[Stack Files]');

    # sort the tracksref so the user is not required to pre-sort them.
    # the top file (index 0) will be the main file; the dummy files 1..n will follow.
    my @tracks = sort by_filename @$tracksref;
    $tracksref = \@tracks;

    for (@pushups, @pushdowns, @required_fields) {
	my ($field, $expr) = split /\|/, $_;
	exists $tracksref->[0]->{lc $field} or 
	    die 'The file list requires ', $expr ?
		"an expression field named '$field' defined with the expression '$expr'" :
		"the '$_' field", "\nStopped";
    }

    $top_filename = $tracksref->[0]->{'filename'};

    for (@pushups) {
	my $lckey = lc $_;
	$tracksref->[1]->{$lckey} ne '' or die "field '$_' is empty in the dummy files and must have a value\nStopped";
	$pushup_fields{$_} = $tracksref->[1]->{$lckey};
    }
    for (@pushdowns) {
	my $lckey = lc $_;
	$pushdown_fields{$_} = $tracksref->[0]->{$lckey} if defined $tracksref->[0]->{$lckey};
    }

    for my $row (@$tracksref) {
	push @filekeys, $row->{'filekey'};
    }

    return $tracksref;
}

sub do_perfile {
    my ($argref, $field, $row) = @_;

    $nfiles++;
    if ($row->{'filekey'} eq $filekeys[0]) {	# top file
	return { 
	    %pushup_fields,
	    'Stack Top'   => $filekeys[0],
	    'Stack Files' => join('|', @filekeys[1..$#filekeys]),
	    'Filename'    => $top_filename . 'tmp'
	};
    }
    else {
	#sleep 5 if $nfiles == 2;
	unlink $row->{'filename'} or die "Failed to unlink file: $row->{'filename'}: $!\nStopped";
	return {
	    %pushdown_fields,
	    'Stack Top' => $filekeys[0],
	    'Stack Tag' => 'Particle',
	    'Filename'  => $top_filename
        };
    }
}

sub do_postprocessing {
    $main::mcws->send([ { 'Filekey' => $filekeys[0], 'Filename' => $top_filename } ]);
};

1;
