# Tests the integrity of an media file.  Currently supported formats:
#
#   flac, m4a
#
# Requirements
#   Each file format uses a specific command line verification tool.  You'll only need the tools
#   for the file types to be tested.  Unsupported file types are ignored.  The file types currently
#   supportd, and tools required are:
#
#    flac
#	flac command line executable
#	    Windows: http://sourceforge.net/projects/flac/files/flac-win/
#	    OS X:    http://sourceforge.net/projects/flac/files/flac-darwin/ or use
#	             via MacPorts: port install flac
#    m4a
#	ffmpeg command line executable
#	    All: http://www.ffmpeg.org/download.html
#	    OS X:    above URL or via MacPorts: port install ffmpeg
#
#   Download the required tools, and place/install them whereever you want.  Configure the tool's
#   name and path via the exe_<tool>_path_<platform> variable in this scriplet's section of the
#   config file.  You can place the tool in your PATH and use its simple filename, or specify the
#   file's path.  Your OS' PATH variable will be used to help resolve the location of the tool.
#
#   See also:
#	http://yabb.jriver.com/interact/index.php?topic=88717.0
#	http://yabb.jriver.com/interact/index.php?topic=85990.msg609284#msg609284
#

package TestFile 1.00;

use v5.16;
use strict;
use warnings;

use File::Which;

my %options = (
    var => { 
	exe_flac_path_win	=> 'flac.exe',
	exe_flac_path_mac	=> 'flac',
	exe_ffmpeg_path_win	=> 'ffmpeg.exe',
	exe_ffmpeg_path_mac	=> 'ffmpeg',
    },
);

# Mapping of supported file suffix to executable name (key 'prog') and path (key 'exe').  The
# 'exe' key will be determined in do_preprocesssing.
my %filesuffix_to_exe = (
    flac => {
	prog => 'flac',
    },
    m4a => {
	prog => 'ffmpeg',
    },
);

$DB::single = 1;

main::set_scriptlet_opts(\%options);

our %callbacks = (
    pre     => \&do_preprocessing,
    perfile => \&do_perfile,
    post    => undef,
);

# Pre-processing callback function
sub do_preprocessing {
    my ($argstr, $tracksref, $fieldsref) = @_;

    # Get the list of file types being worked on to verify that the files types are supported for verification
    my (%filetypes, %ignored);
    for (@$tracksref) {
	my ($suffix) = $_->{'filename'} =~ /\.(.+)$/;
	if ($suffix ~~ %filesuffix_to_exe) {
	    $filetypes{$suffix}++;
	    next;
	}
	say "Ignoring unsupported file type ", $suffix	 unless $ignored{$suffix}++;
    }

    # For each file type, check that the corresponding decode program exists and is executable.
    for (keys %filetypes) {
	my $exepath = $options{'var'}{join('_', 'exe', $filesuffix_to_exe{$_}{'prog'}, 'path', $^O eq 'MSWin32' ? 'win' : 'mac')};

	$filesuffix_to_exe{$_}{'exe'} = ($exepath =~ /^(?:\/|[a-z]:\\)/i ? $exepath : which($exepath));
	
	-e $filesuffix_to_exe{$_}{'exe'} or
	    die 'The ', $filesuffix_to_exe{$_}{'prog'}, ' executable was not found at ', $filesuffix_to_exe{$_}{'exe'}, "\nStopped";

	unless ($^O eq 'MSWin32') {
	    -x $filesuffix_to_exe{$_}{'exe'} or
		die $filesuffix_to_exe{$_}{'prog'}, " program is not executable\nStopped";
	}
    }

    return undef;
}

# Per-row callback function
sub do_perfile {
    my ($argstr, $field, $row) = @_;

    # Unicode confounds qx//, so try the ANSI name first, and if that fails, try the original name.  Some
    # files will fail, and fixing this is non-trivial since it requires setting up a different process
    # creation and STDOUT capture mechanism.
    my (@names_to_try);
    if ($^O eq 'MSWin32') {
	use if $^O eq 'MSWin32', 'Win32';

	if (my $ansi_name = Win32::GetANSIPathName($row->{'filename'})) {
	    push @names_to_try, $ansi_name;
	}
	else {
	    warn "Win32::GetANSIPathName unable to get ANSI path name: $^E.\nStopped";
	}
    }

    my $i = 0;
    my $response;
    for (@names_to_try, $row->{'filename'}) {
	main::debug "\tTrying filename(", ++$i, '): ', $_;
	my ($suffix) = $_ =~ /\.(.+)$/;
	if ($suffix =~ /^flac$/) {
	    $response = qx("$filesuffix_to_exe{$suffix}{'exe'}" -st "$_" 2>&1);
	    last if $response eq '';
	    next if $response =~ /FLAC__STREAM_DECODER_INIT_STATUS_ERROR_OPENING_FILE/ms;
	    $response =~ s/^\R+//;
	    $response =~ s/\R+$//;
	    $response =~ s/^[^:]+: //ms;
	    last;
	}
	elsif ($suffix =~ /^m4a$/) {
	    $response = qx("$filesuffix_to_exe{$suffix}{'exe'}" -v error -i "$_" -f null - 2>&1);
	    chomp $response;
	    last if $response eq '' or $response =~ /Error while decoding stream/;
	    next if $response =~ /No such file or directory/;
	}
	else {
	    main::verbose 'Ignoring unsupported ', $suffix, ' file type: ', $_;
	    next;
	}

	main::debug "$filesuffix_to_exe{$suffix}{'prog'} file($i): ", $_, ': ', $response;
    }

    return undef unless $response;
}

1;
