# Sample complex script template showing how to perform work during and after all rows have been seen
# This is a sample scriptlet showing how to do pre- and post-processing.  It can be used as a general
# tempalte for create your own script for use with pscriptor.pl.
#
# Your script will be called by pscriptor.pl as follows:
#
#	1. pre-processing
#	2. per-file processing
#	3. post-processing
#
# See the %callbacks hash comments below, and the comments above each sample funciont.
# Any code placed outside these functions is run once, when pscriptor loads the modules.
# This is where to place any required initiazation code.
#

package SampleComplete 1.01;

use v5.16;
use strict;
use warnings;

# The %callbacks hash table defines the callback functions called during in the stages mentioned above.
# Set a function to undef if you do not need it to be called.
our %callbacks = (
    pre     => \&do_preprocessing,
    perfile => \&do_perfile,
    post    => \&do_postprocessing,
);

# This will cause the perl debugger to stop here when debugging and is harmless otherwise.  Without it,
# you'd have to debug pscriptor itself and single-step into this module.
$DB::single = 1;

my $rows_seen = 0;

# Pre-processing callback function
#
# Do your pre-file work here.  In this stage, you have access to the entire table of MC view data,
# each row is a file's metadata in the view, and each column is a header in the view.
#
# Input: The pre-processsing routine will receive three arguments:
#
#      1. An argument passed via --args|-a, used to possibly control this scriptlet
#
#      2. A reference to the entire table of the selected tracks, each row a hash of field -> values.
#
#      3. A list of all the input fields passed via -f
#
# Ouput: undef, or a reference to a modified tracksref (i.e. maybe to sort it, or delete rows)
#
sub do_preprocessing {
    my ($argstr, $tracksref, $fieldsref) = @_;

    say "Pre-processing: called, with ", scalar @{$tracksref}, "rows of data and argument ", $argstr;

    return undef;
}

# Per-row callback function
#
# Do your per-file work here.  This is thep only place where cell values can be modified.
#
# Input: The per-file routine will receive three arguments:
#
#      1. An argument passed via --args|-a, used to possibly control this scriptlet
#
#      2. The name of the current field being opearted on.
#
#      3. A hash that refers to the current row of cells in the MC view.  Each column name in the file list
#         is a key, and the hash can be used to access MCs' cell values.  The keys are all forced to lowercase,
#         so be sure to user lowercase keys when using the %file hash.  For example, use the string 'filename'
#         in $row->{'filename'}, which will contain MC's full path to the file currently being operated on.  If
#         the -f option specified the 'Name' field, then the value $row->{$field} will be exactly the same as
#         $row->{'name'}.
#
# Output: Setting the return value to something other than the cell's current value will cause the cell to be
#         updated with the return value.  If the return value is undefined, empty or unchanged, the cell data
#	  will be left unmodified.
#
sub do_perfile {
    my ($argstr, $cellval, $row) = @_;

    # per-row code
    # $row is a hash reference to the current row.  The hash keys are the view headers, in lowercase.
    # For example, $row->{'filename'} is the value of the Filename field.
    $rows_seen++;
    say "\t\tPer-row call($rows_seen) call with field value '$row->{$field}'; File is $row->{'filename'}";

    return undef;
}

# Post-processing callback function
#
# Do any postporocessing work here.  This function will be called after all per-file processing has been completed.
# This is where you can do any post-processing work after all per-file calculations have been made.
#
# Input: The post-processing routine will receive one argument:
#
#      1. An argument passed via --args|-a, used to possibly control this scriptlet
#
# Ouput: none
#
sub do_postprocessing {
    my $argstr = shift;

    say "Post-processing callback; row_seen = $rows_seen";
};

1;
