#!/usr/bin/perl

use v5.14;
use utf8;
use strict;
use warnings;
use diagnostics;

#binmode STDOUT, ':encoding(UTF-8)';
#binmode STDERR, ':encoding(UTF-8)';

my $version = "1.15pre3";

use LWP::UserAgent;
use Getopt::Long;
use File::Basename;
use Time::HiRes qw(usleep);
use Tie::File;
use Text::CSV;
use URI::Escape;
use XML::LibXML '1.70';
use if $^O eq 'MSWin32', 'Win32::Clipboard' => 'CF_UNICODETEXT';
use if $^O eq 'darwin', 'Mac::Pasteboard';
use MCUtils::MCWS;
use MCUtils::Misc;
#use Data::Dumper;

# Look in the Scriptlets directory for modules
use lib "Scriptlets";

my $prog = basename($0);
my $progstr = "$prog version $version";
my $default_config_file = basename($prog, '.pl') . '-config.txt';

sub Usage {
    my $exitcode = shift;
    say @_ ? join('', @_, "\n") : '',
    <<ENDUSAGE, "\nStopped";
Usage: $prog <options>
    options:
    --args       | -a <scriptlet args>		# argument string passed to scriptlet
    --configfile | -c <config file>		# read options from config file upon startup
    --debug      | -d				# enable debug output
    --escript    | -e				# executable perl code
    --efile      | -E <module name>		# executable perl code from specified module
    --field      | -f				# field(s) to change
    --filekey    |                              # use FileKey() value instead of Filename (required for particles, special MC entries)
    --help	 | -h				# output help and usage text
    --mpl					# output MPL
    --nomcws					# don't use mcws service
    --outfile    | -o <output file.mpl>		# use file specified as the MPL output file
    --password   | -P <username:password>	# MCWS username and password
    --server     | -S <server:port>		# MCWS server and port
    --testonly   | -t              		# test-only with mcws: no updates performed (implies -v)
    --verbose    | -v				# output operations more verbosely
ENDUSAGE
    exit $exitcode;
}

our %opts = ();
our $mcws;

our %scriptlet_opts;	# used to pass config file options to scriptlets
our %mcwsparams;	# MCWS http connection parameters
my %callbacks;

# Read the default config file.
read_config_file($default_config_file) if -e $default_config_file;
my @save_ARGV = @ARGV;

{
local $SIG{__WARN__} = sub { say "\n*** ", $_[0]; };

Getopt::Long::Configure('no_ignore_case');
GetOptions(\%opts,
	'args|a=s',
	'configfile|c=s' => sub { read_config_file($_[1]) },
	'debug|d'	 => sub { debug_on() },
	'escript|e=s',
	'efile|E=s',
	'field|f=s@',
	'filekey',
	'help|h'         => sub { Usage(0) },
	'mpl',
	'nomcws',
	'outfile|o=s',
	'password|P=s'	 => sub { ($mcwsparams{'username'}, $mcwsparams{'password'}) = split(/:/, $_[1], 2) },
	'server|S=s'	 => sub { $mcwsparams{'server'} = $_[1] },
	'testonly|t'     => sub { verbose_on(); $opts{$_[0]} = 1 },
	'verbose|v'	 => sub { verbose_on() },
    ) or Usage(1);
}
debug "Command Line: @save_ARGV";

Usage(1, "Must use exactly one of --escript or --efile.  Expanded command line was:\n\n\t@save_ARGV\n") unless exists $opts{'efile'} xor exists $opts{'escript'};
Usage(1, 'The --field option is required to specify the field(s) that should be updated.') unless exists $opts{'field'};

#exists $opts{'testonly'} and verbose_on();	# force verbose when in test mode

# canonicalize field names to lowercase.
foreach (@{$opts{'field'}}) {
    $_ = lc;
}

# setup the connection to MCWS if not disabled
unless (exists $opts{'nomcws'}) {
    $mcws = MCUtils::MCWS->connect(%mcwsparams);

    $mcws->do_mcc_command(21000);		# Copy the clipboard for the user
}

my $cliptracks = [];

$cliptracks = parse_clipboard( get_clipboard_contents() );

@$cliptracks > 0 or
    die "Clipboard does not appear to have copied track data.\nSelect the files in MC and ", 
	exists $opts{'nomcws'} ? 'Copy' : 'they will be copied automatically.',
	"\nOr use the --create option to create dummy files\nStopped";

my $filefield = $opts{'filekey'} ? 'filekey' : 'filename';
/^$filefield$/i ~~ %{$cliptracks->[0]} or
    die "Clipboard does not have the $filefield column, or clipboard data is not from MC.\nStopped";

# read the script from a file if necessary
if ($opts{'efile'}) {
    require "$opts{'efile'}.pm";
    #$opts{'efile'} =~ s/^.*?([^\\\/]+)$/$1/g;
}

# Setup the default output file.
if ($^O eq 'MSWin32') {
    defined $opts{'outfile'} or
	$opts{'outfile'} = join('\\', $ENV{'HOMEPATH'}, 'Desktop', 'import.mpl');
}
else {
    defined $opts{'outfile'} or
	$opts{'outfile'} = join('/', $ENV{'HOME'}, 'Desktop', 'import.mpl');
}


# ensure the -f specified fields are in the view
foreach (@{$opts{'field'}}) {
    my $field = (split /::/)[0];
    /^\Q$field\E$/ ~~ %{$cliptracks->[0]} or
	die "Field '$field' not available in the view.  Check field specified in -f and the view"
}

# Call the pre-processing callback, giving the scriptlet a chance ot evaluate all the data in the view.
if (exists $opts{'efile'}) {
    no strict "refs";

    my $cb = \%{$opts{'efile'} . '::callbacks'};
    $callbacks{'pre'}     = (exists $cb->{'pre'}     and defined $cb->{'pre'})     ? $cb->{'pre'}     : undef;
    $callbacks{'perfile'} = (exists $cb->{'perfile'} and defined $cb->{'perfile'}) ? $cb->{'perfile'} : undef;
    $callbacks{'post'}    = (exists $cb->{'post'}    and defined $cb->{'post'})    ? $cb->{'post'}    : undef;

    # Arguments: scriptlet args, ref to entire view table, and the ref to list of input fields.
    my @fields = map { (split(/::/, $_))[0] } @{$opts{'field'}};
    if ($callbacks{'pre'}) {
	if (my $ret = $callbacks{'pre'}->($opts{'args'}, $cliptracks, \@fields)) {
	    $cliptracks = $ret;		# allow scriptlets to modify the tracklist
	}
    }
}

if (my $outlist_ref = process_user_tracks($cliptracks)) {
    output_mpl($outlist_ref)	if $opts{'mpl'};
    $mcws->send($outlist_ref)	if ! exists $opts{'nomcws'} and ! exists $opts{'testonly'};
}

# Call the post-processing callback, giving the scriptlet a chance to do any work after all files have been processed.
if ($callbacks{'post'}) {
    $callbacks{'post'}->($opts{'args'})
}

# cycle through the user's tracks on the clipboard
sub process_user_tracks {
    my $clipsref = shift;
    my @output_list = ();

    my $i = 1;

    for my $file (@$clipsref) {
	my %h = ();
	unless (exists $opts{'efile'} and !$callbacks{'perfile'}) {
	    verbose 'File: ', $opts{'filekey'} ?
	                        $file->{'filekey'} . ($file->{'filename'} ? ': ' . $file->{'filename'} : '') :
				$file->{'filename'};
	}
	for my $fieldspec (@{$opts{'field'}}) {
	    my ($fieldin, $fieldout) = split /::/, $fieldspec;

	    $fieldout = $fieldin	 unless $fieldout;

	    # Scriptlet code
	    if (exists $opts{'efile'}) {
		# Call the module's per-file callback code
		next unless $callbacks{'perfile'};
		$_ = $callbacks{'perfile'}->($opts{'args'}, $fieldin, $file);
	    }

	    # Simple perl script code is just eval'ed in the current namespace
	    else {
		no strict;
		$_ = $file->{$fieldin} || '';
		eval $opts{'escript'};
		die "The expression:\n\n\t$opts{'escript'}\n\nfailed to execute: $@\nStopped" if $@;
	    }

	    verbose "\tin($fieldin): '$file->{$fieldin}'";
	    # HASH return: the scriptlet defines the key / value pairs for the fields to be updated.
	    if (ref $_ eq 'HASH') {
		for my $key (keys %$_) {
		    verbose "\tnew($key): ", $_->{$key};
		    $h{$key} = $_->{$key}  if $_->{$key} ne '';
		}
	    }

	    # SCALAR return
	    elsif (ref $_ eq '') {		# SCALAR
		my $new = defined $_ ? $_ : '';
		my $old = exists $file->{$fieldout} ? $file->{$fieldout} : '';
		if ($new ne $old and $new ne '') {
		    verbose "\told($fieldout): ", ($fieldout ne $fieldin and exists $file->{$fieldout}) ? "'$file->{$fieldout}'" : '<unknown>';
		    verbose "\tnew(", $fieldout , "): '$new'";
		    $h{$fieldout} = $new;
		}
		else {
		    verbose "\told=new($fieldout): ", ($fieldout ne $fieldin and exists $file->{$fieldout}) ? "'$file->{$fieldout}'" : '<unknown>',
		            "\tfield($fieldout) UNCHANGED: ", $new eq '' ? 'null value returned from script' : 'new == old' ;
		}
	    }
	    else {
		die "Unknown return type from scriptlet: ", ref $_, "\nStopped";
	    }
	}
	if (%h) {
	    if ($opts{'filekey'}) {
		$h{'filekey'} = $file->{'filekey'};
	    }
	    else {
		$h{'Filename'} = $file->{'filename'};
	    }
	    push @output_list, \%h;
	    debug 'Added to output list: : ', $opts{'filekey'} ?
						$file->{'filekey'} . ($file->{'filename'} ? ': ' . $file->{'filename'} : '') :
						$file->{'filename'};
	}
    }

    return \@output_list;
    exit;
}

my @config_stack;
# read and process a config file
sub read_config_file {
    my $configfile = shift;

    push @config_stack, $configfile;
    open my $io, "<", $configfile or
	die "can't open config file $configfile\n$!\nStopped";

    while (<$io>) {
	s/\R+$//;
	next if /^\s*$/;
	next if /^\s*#/;

	# Options:  command line options
	if (/Options:\s*(?<opts>.+)$/) {
	    unshift @ARGV, split /\s+/, $+{'opts'};
	    debug "Options: $+{'opts'}\n";
	}

	# <scriptlet>_<option>:  <key> = <value>
	elsif (/(?<scriptlet>\S+)::(?<option>[^:]+):\s*(?<key>\S+)\s*=\s*(?<value>.+)$/) {
	    my ($scriptlet, $option, $key, $value) = ($+{scriptlet}, $+{option}, $+{key}, $+{value});
	    $value =~ s/#.*$//; $value =~ s/\s*$//;
	    $option = lc $option; $key = lc $key;
	    debug "scriptlet $scriptlet option $option: $key ==> '$value'";
	    $scriptlet_opts{$scriptlet}{$option}{$key} = $value;
	}

	elsif (/MCWS_(server|username|password):\s*(.+?)\s*$/) {
	    $mcwsparams{$1} = $2;
	}

	# include: pscriptor-config-custom.txt
	elsif (/^include:\s*(.+?)\s*$/) {
	    die "Too many config file inclusions... don't have files include each other.\nConfig file inclusions:\n\t",
		join("\n\t", @config_stack), "\nStopped"   if @config_stack > 10;
	    if (-e $1) {
		read_config_file($1);
	    }
	    else {
		warn "Included config file does not exist: $1\n";
	    }
	}
    }
    close $io;
    pop @config_stack;
}

# Sets the scriptlet options from those found in the config file specific to the scriptlet.
sub set_scriptlet_opts {
    my $optref = shift;

    if (exists $scriptlet_opts{$opts{'efile'}}) {
	my $h = $scriptlet_opts{$opts{'efile'}};
	for my $type (keys %$h) {
	    unless ($type ~~ %$optref) {
		say "Ignoring unknown scriptlet option type '$type'";
		next;
	    }
	    for my $key (keys $h->{$type}) {
		unless ($key ~~ %{$optref->{$type}}) {
		    say "Ignoring unknown scriptlet $type '$key'";
		    next;
		}
		$optref->{$type}{$key} = $h->{$type}{$key};
	    }
	}
    }
}

sub open_file {
    my $filepath = shift;
    my $mpl;

    if ($filepath eq '-') {
	$mpl = \*STDOUT;
	delete $opts{'append'};		# ensure writing of the MPL preamble
	return $mpl;
    }
    else {
	my $openmode = '>';

	if (exists $opts{'append'}) {
	    if (-e $filepath and ! -z _) {
		my @lines;

		tie(@lines, 'Tie::File', $filepath) or
		    die "Can't update file $filepath\nStopped";
		delete $lines[-1] if $lines[-1] =~ /^<\/MPL>/;
		untie @lines;
		$openmode = '>>';
	    }
	    else {
		delete $opts{'append'};	# ensure writing of the MPL preamble
	    }
	}

	#my $encoding = ' :utf8';
	my $encoding = '';
	open($mpl, $openmode . $encoding, $filepath) or
	    die "Failed to open $filepath: $!\nStopped";
    }

    return $mpl;
}

sub output_mpl {
    my $lref = shift;
    return unless $lref;

    sub create_item {
	my $cellsref = shift;
	my $str;
	my $count = 0;

	for (keys %$$cellsref) {
	    if ($$cellsref->{$_} and $$cellsref->{$_} ne '' and $$cellsref->{$_} ne '-') {
		my $val = $$cellsref->{$_};
		$val =~ s/\&(?!amp;|quot;)/&amp;/g;
		$val =~ s/</&lt;/g;
		$val =~ s/>/&gt;/g;
		$str .= '<Field Name="' . $_ . '">'. $val . "</Field>\n";
		$count++;
	    }
	}
	unless ($count) {
	    verbose 'MPL: Nothing to update for file: ', $$cellsref->{'filename'};
	    return '';
	}

	return "<Item>\n" . $str . "</Item>\n";
    }

    my $s;
    for (@$lref) {
	$s .= create_item(\$_);
    }

    if ($s) {
	my $mpl = open_file($opts{'outfile'});

	print $mpl qq/<?xml version="0.0" encoding="UTF-8" standalone="yes" ?>\n/,
		   qq/<MPL Version="2.0" Title="Created from $progstr">\n/;

	print $mpl $s, "</MPL>\n";
    }
}

1;
